/*
** PHREL
** $Id: thread.c,v 1.56 2013/06/01 04:55:31 sella Exp $
** Copyright (c) 2005-2013 James M. Sella. All Rights Reserved.
** Released under the GPL Version 2 License.
** http://www.digitalgenesis.com
*/

static const char rcsid[] __attribute__((used)) = "$Id: thread.c,v 1.56 2013/06/01 04:55:31 sella Exp $";

#include "thread.h"

#include "db.h"
#include "trap.h"
#include "util.h"
#include "phrel.h"
#include "prefs.h"
#include "data_hash.h"
#include "capture.h"

#include <snmptrap.h>

#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <sys/wait.h>
#include <sys/types.h>

#define SAFE_MYSQL_FREE_RESULT(x) if (x) { mysql_free_result(x); x = NULL; }

/* Global variables */
struct data_list_t *data_list = NULL;
struct data_list_t **dl_end = &data_list;
uint32_t d_count = 0;

/* External Global Variables */
extern struct args_t args;
extern int running, restart;
extern pid_t pid;
extern time_t startup_time;
extern pthread_mutex_t data_mutex;
extern pthread_mutex_t db_mutex;
extern struct hash_key_t hash_key[HASH_WIDTH];
#ifdef HAVE_MYSQL
extern MYSQL mysql;
#endif

void data_thread(void) {
	char buf[1024], buf2[46];
	time_t next, stat_next = 0;
	unsigned int i, pps, count = 0;
	struct data_t *d_ptr;
	struct hash_entry_t **entry_h = NULL;
	FILE *fd = NULL;
#ifdef HAVE_MYSQL
	int connected;
#endif

	/* Insert sync chain */
	chain_sync_add();

	while (running) {
		next = time(NULL) + args.check_interval;
		while (running && next > time(NULL)) {
			usleep(750000);
		}

		/* Verify sync chain exists. */
		if (chain_sync_check() == 0) {
			syslog(LOG_INFO, "out of sync with iptables - restarting");

			snmptrap_sync_lost();

			running = 0;
			restart = 1;

			continue;
		} else if (args.debug >= DEBUG_INTERNAL) {
			syslog(LOG_INFO, "in sync with iptables");
		}

		/* Open stats file, if needed. */
		if (args.stats != NULL && time(NULL) > stat_next) {
			stat_next = time(NULL) + args.stats_interval;
			if (args.debug >= DEBUG_INTERNAL) {
				syslog(LOG_INFO, "opening stats file %s", args.stats);
			}

			if ((fd = fopen(args.stats, "w")) == NULL) {
				syslog(LOG_INFO, "failed to open stats file %s - disabling stats", args.stats);

				/* Disable stats file. */
				free(args.stats);
				args.stats = NULL;
			} else if (hash_stats(buf, sizeof(buf)) == 0) {
				if (fwrite(buf, 1, strlen(buf), fd) == 0) {
					syslog(LOG_INFO, "failed write to stats file %s - closing file", args.stats);
					/* Write failed, close stats file. */
					fclose(fd);
					fd = NULL;
				}
			}
		}

		if (args.debug >= DEBUG_INTERNAL) {
			syslog(LOG_INFO, "data thread starting to walk hash...");
		}

		for (i = 0; i < HASH_WIDTH && running; i++) {
			pthread_mutex_lock(&data_mutex);

			if ((entry_h = &hash_key[i].entry) != NULL) {
				while (*entry_h != NULL) {
					d_ptr = (*entry_h)->data;

					pps = calc_pps(d_ptr);
					process_threshold(d_ptr, pps);

#ifdef HAVE_MYSQL 
					/* Add data to data_list. */
					if (d_count > HASH_WIDTH) {
						if (args.debug >= DEBUG_ERROR) {
							inet_ntop(d_ptr->ss.ss_family, get_in_addr(&(d_ptr->ss)), buf2, sizeof(buf2));
							syslog(LOG_WARNING, "Maximum data queue depth reached. Tail dropping new entry for '%s'", buf2);
						}
					} else if (d_ptr->local.peak >= args.db_threshold) {
						pthread_mutex_lock(&db_mutex);
						connected = db_isconnected();
						pthread_mutex_unlock(&db_mutex);

						if (connected && d_ptr->local.last_update > 0) {
							*dl_end = malloc(sizeof(struct data_list_t));
							memcpy(&(*dl_end)->ss, &(d_ptr->ss), sizeof(struct sockaddr_storage));
							(*dl_end)->updated = d_ptr->local.last_update;
							(*dl_end)->avg = d_ptr->local.avg;
							(*dl_end)->sum = d_ptr->local.sum;
							(*dl_end)->max = d_ptr->local.max;

							(*dl_end)->next = NULL;
							dl_end = &(*dl_end)->next;
							d_count++;
						}
					}
#endif

					if (args.debug >= DEBUG_DUMP2) {
						inet_ntop(d_ptr->ss.ss_family, get_in_addr(&(d_ptr->ss)), buf2, sizeof(buf2));
						syslog(LOG_INFO, "checking if %s is stale", buf2);
					}

					if (fd != NULL) {
						inet_ntop(d_ptr->ss.ss_family, get_in_addr(&(d_ptr->ss)), buf2, sizeof(buf2));
						snprintf(buf, sizeof(buf), "%s\t%8u pps %8u threshold %8u rate\n", buf2, pps, d_ptr->threshold, d_ptr->rate);
						if (fwrite(buf, 1, strlen(buf), fd) < strlen(buf)) {
							syslog(LOG_ERR, "short write detected to stat file");
						}
					}

					/* Move to next item. */
					entry_h = &(*entry_h)->next;

					if (check_stale(d_ptr) == 1) {
						if (args.debug >= DEBUG_DUMP1) {
							inet_ntop(d_ptr->ss.ss_family, get_in_addr(&(d_ptr->ss)), buf2, sizeof(buf2));
							syslog(LOG_INFO, "removing %s from hash (stale)", buf2);
						}

						hash_del(&(d_ptr->ss)); /* Invalidates (*entry_h). */

						/* Reset (*entry_h), since hash_del invalidates the pointer. */
						entry_h = &hash_key[i].entry;
					}
				}
			}

			pthread_mutex_unlock(&data_mutex);

			/* Don't peg CPU. Loop doesn't need to be completed super quickly. */
			if ((count++ % (HASH_WIDTH / 7)) == 0) { /* 7 times per cycle. */
#ifdef _POSIX_PRIORITY_SCHEDULING
	   		sched_yield();
#else
				usleep(0);
#endif
			}
		}

		if (args.debug >= DEBUG_INTERNAL) {
			syslog(LOG_INFO, "data thread completed walking hash");
		}

		/* Close stats file. */
		if (fd != NULL) {
			if (args.debug >= DEBUG_INTERNAL) {
				syslog(LOG_INFO, "closing stats file %s", args.stats);
			}

			fclose(fd);
			fd = NULL;
		}
	}

	/* Delete sync chain */
	chain_sync_del();

	if (args.debug >= DEBUG_INTERNAL) {
		syslog(LOG_INFO, "exiting data thread");
	}

	pthread_exit(NULL);
}

#ifdef HAVE_MYSQL 
void db_thread(void) {
	time_t next;

	struct data_list_t *dl_ptr;

	while (running) {
		next = time(NULL) + args.check_interval;
		while (running && next > time(NULL)) {
			/* Write pending data */
			while (data_list) {
				/* Grab reference to current data. */
				dl_ptr = data_list;

				/* Write the data. */
				if (write_data(dl_ptr) != 0) {
					db_reconnect();
				}

				/* Decrease data count by 1 */
				d_count--;

				/* Move to the next data. */
				if ((data_list = data_list->next) == NULL) {
					dl_end = &data_list;
				}

				/* Free the node. */
				free(dl_ptr);
				dl_ptr = NULL;
			}

			usleep(750000);
		}

		if (clean_data() != 0 || read_data() != 0) {
			db_reconnect();
		}
	}

	if (args.debug >= DEBUG_INTERNAL) {
		syslog(LOG_INFO, "exiting db thread");
	}

	pthread_exit(NULL);
}

int db_reconnect(void) {
	time_t next = 0;
	int err = 0;
	int connected = 0;

	pthread_mutex_lock(&db_mutex);
	db_disconnect();
	pthread_mutex_unlock(&db_mutex);

	while (running && connected == 0) {
		while (running && connected == 0 && next > time(NULL)) {
	 		usleep(750000);
		}
		next = time(NULL) + DEF_DBTIMO;

		if (args.debug >= DEBUG_ERROR) {
			syslog(LOG_INFO, "attempting to reconnect to MySQL server");
		}

		pthread_mutex_lock(&db_mutex);
		err = db_connect();
		connected = db_isconnected();
		pthread_mutex_unlock(&db_mutex);

		if (connected != 0 && args.debug >= DEBUG_ERROR) {
			syslog(LOG_INFO, "reconnect to MySQL server");
		}
	}

	return err;
}

int read_data() {
	MYSQL_RES *res = NULL;
	MYSQL_ROW row;
	char sql[1024];
	char buf[46];
	struct data_t d, *d_ptr;
	struct sockaddr_storage ss;

	switch (args.db_algo) {
		case PPS_ALGO_AVG:
			snprintf(sql, sizeof(sql), "SELECT ip, MAX(updated) - (UNIX_TIMESTAMP() - %ld) updated, AVG(avg) avg, AVG(sum) sum, AVG(max) max FROM data WHERE client != %ld AND updated > (UNIX_TIMESTAMP() - %u) GROUP BY ip", time(NULL), mysql_thread_id(&mysql), args.stale_interval);
			break;

		case PPS_ALGO_SUM:
			snprintf(sql, sizeof(sql), "SELECT ip, MAX(updated) - (UNIX_TIMESTAMP() - %ld) updated, SUM(avg) avg, SUM(sum) sum, SUM(max) max FROM data WHERE client != %ld AND updated > (UNIX_TIMESTAMP() - %u) GROUP BY ip", time(NULL), mysql_thread_id(&mysql), args.stale_interval);
			break;

		default:
		case PPS_ALGO_MAX:
			snprintf(sql, sizeof(sql), "SELECT ip, MAX(updated) - (UNIX_TIMESTAMP() - %ld) updated, MAX(avg) avg, MAX(sum) sum, MAX(max) max FROM data WHERE client != %ld AND updated > (UNIX_TIMESTAMP() - %u) GROUP BY ip", time(NULL), mysql_thread_id(&mysql), args.stale_interval);
			break;
	}

	pthread_mutex_lock(&db_mutex);

   if (db_query(&res, sql) != 0) {
		pthread_mutex_unlock(&db_mutex);

      return 1;
   }

	pthread_mutex_unlock(&db_mutex);

   if (res != NULL && mysql_num_rows(res) < 1) {
		SAFE_MYSQL_FREE_RESULT(res);

		return 0;
	}

	while (res != NULL && (row = mysql_fetch_row(res))) {
		if (strchr(row[0], ':') == NULL) {
			ss.ss_family = AF_INET;
			inet_pton(AF_INET, row[0], get_in_addr(&ss));
		} else {
			ss.ss_family = AF_INET6;
			inet_pton(AF_INET6, row[0], get_in_addr(&ss));
		}

		if (args.debug >= DEBUG_SQL) {
			syslog(LOG_INFO, "SQL - ip: %s, updated: %s (%ld sec ago), avg: %s, sum: %s, max: %s", row[0], row[1], time(NULL) - strtoul(row[1], NULL, 10), row[2], row[3], row[4]);
		}

		pthread_mutex_lock(&data_mutex);

		if ((d_ptr = hash_find(&ss)) == NULL) {
			/* Check Filters - Do we track this sources? */
			if (check_filter(&ss) != 0) {
				pthread_mutex_unlock(&data_mutex);

				continue; /* skip source */
			}

			/* Build new data_t structure */
			d_ptr = &d;
			memset(d_ptr, 0, sizeof(struct data_t));
			d_ptr->ss = ss;

			/* Dynamically allocate samples */
			d_ptr->samples = malloc((args.interval + 1) * sizeof(unsigned int));
			memset(d_ptr->samples, 0, (args.interval + 1) * sizeof(unsigned int));

			/* Dynamically allocate last_update */
			d_ptr->last_update = malloc((args.interval + 1) * sizeof(time_t));
			memset(d_ptr->last_update, 0, (args.interval + 1) * sizeof(time_t));

			if (args.debug >= DEBUG_DUMP1) {
				inet_ntop(ss.ss_family, get_in_addr(&ss), buf, sizeof(buf));
				syslog(LOG_INFO, "db adding %s to hash", buf);
			}

			hash_add(&d_ptr->ss, d_ptr);
		}

		d_ptr->db.last_update = strtoul(row[2], NULL, 10);
		d_ptr->db.avg = strtoul(row[3], NULL, 10);
		d_ptr->db.sum = strtoul(row[4], NULL, 10);
		d_ptr->db.max = strtoul(row[5], NULL, 10);

		pthread_mutex_unlock(&data_mutex);
	}

	SAFE_MYSQL_FREE_RESULT(res);

	return 0;
}

int write_data(struct data_list_t *dl_ptr) {
	int ret_val = 1;
	MYSQL_RES *res = NULL;
	char sql[1024];
	char buf[1024];

	pthread_mutex_lock(&db_mutex);

	if (inet_ntop(dl_ptr->ss.ss_family, get_in_addr(&(dl_ptr->ss)), buf, sizeof(buf)) == NULL) {
		goto exit;
	}

	snprintf(sql, sizeof(sql), "UPDATE data SET updated = UNIX_TIMESTAMP() - %ld, avg = %u, sum = %u, max = %u WHERE client = %ld AND ip = '%s'", time(NULL) - dl_ptr->updated, dl_ptr->avg, dl_ptr->sum, dl_ptr->max, mysql_thread_id(&mysql), buf);

   if (db_query(&res, sql) != 0) {
      goto exit;
   }

   if (mysql_affected_rows(&mysql) < 1) {
		SAFE_MYSQL_FREE_RESULT(res);

		snprintf(sql, sizeof(sql), "INSERT INTO data (client, ip, updated, avg, sum, max) VALUES (%ld, '%s', UNIX_TIMESTAMP() - %ld, %u, %u, %u)", mysql_thread_id(&mysql), buf, time(NULL) - dl_ptr->updated, dl_ptr->avg, dl_ptr->sum, dl_ptr->max);

		if (db_query(&res, sql) != 0) {
			goto exit;
		}
	}

	SAFE_MYSQL_FREE_RESULT(res);

	/* Success */
	ret_val = 0;
exit:
	pthread_mutex_unlock(&db_mutex);

	return ret_val;
}

int clean_data() {
	int ret_val = 1;
	MYSQL_RES *res = NULL;
	char sql[1024];

	pthread_mutex_lock(&db_mutex);

	if (running) {
		snprintf(sql, sizeof(sql), "DELETE FROM data WHERE (client = %ld AND updated < UNIX_TIMESTAMP() - %u) OR updated < UNIX_TIMESTAMP() - %u", mysql_thread_id(&mysql), args.stale_interval, args.stale_interval * STALE_INTERVAL_MLT);
	} else {
		snprintf(sql, sizeof(sql), "DELETE FROM data WHERE (client = %ld)", mysql_thread_id(&mysql));
	}

   if (db_query(&res, sql) != 0) {
      goto exit;
   }

	if (args.debug >= DEBUG_SQL) {
		syslog(LOG_INFO, "SQL - deleted %llu rows", mysql_affected_rows(&mysql));
	}

	SAFE_MYSQL_FREE_RESULT(res);

	/* Success */
	ret_val = 0;
exit:
	pthread_mutex_unlock(&db_mutex);

	return ret_val;
}
#endif

void cleanup(void) {
	int i;
	struct data_t *d_ptr;
	struct hash_entry_t **entry_h = NULL;

	/* Cleanup for shutdown. Stop filtering. */
	if (args.debug >= DEBUG_INTERNAL) {
		syslog(LOG_INFO, "cleaning up iptables...");
	}

	pthread_mutex_lock(&data_mutex);
	for (i = 0; i < HASH_WIDTH; i++) {
		if ((entry_h = &hash_key[i].entry) != NULL) {
			while (*entry_h != NULL) {
				d_ptr = (*entry_h)->data;

				if (d_ptr->threshold > 0) {
					chain_del(d_ptr);
				}

				entry_h = &(*entry_h)->next;
			}
		}

	}
	pthread_mutex_unlock(&data_mutex);

	if (args.debug >= DEBUG_INTERNAL) {
		syslog(LOG_INFO, "completed cleanup");
	}
}

int check_stale(struct data_t *data) {
	time_t cur;
	unsigned int i;

	/* If there is a threshold set, its not stale. */
	if (data->threshold > 0) {
		return 0;
	}

	cur = time(NULL);

	/* If newly updated from database, don't remove */
	if (data->db.last_update + args.stale_interval >= cur) {
		return 0;
	}

	for (i = 0; i < args.max_samples; i++) {
		/* If recently updated, don't remove. */
		if (data->last_update[i] + args.stale_interval >= cur && data->last_update[i] != 0) {
			return 0;
		}
	}

	return 1;
}

void process_threshold(struct data_t* data, unsigned int pps) {
	char buf[46], buf2[46];
	struct threshold_t *t;
	time_t cur = time(NULL);
	unsigned int threshold, prev_threshold;
	unsigned int rate, prev_rate;
	unsigned int burst, prev_burst;
	unsigned int decay, prev_decay;

	/* Capture previous threshold, rate, burst and decay. */
	prev_threshold = data->threshold;
	prev_rate = data->rate;
	prev_burst = data->burst;
	prev_decay = data->decay;

	/* Clear threshold, rate and decay if decay has expired. */
	if (data->threshold > 0 && cur > data->decay) {
		data->threshold = 0;
		data->rate = 0;
		data->burst = 0;
		data->decay = 0;
	}

	/* Check for new threshold crossings. */
	t = args.threshold;
	while (t != NULL) {
		if (pps >= t->threshold && (t->threshold >= data->threshold || cur > data->decay)) {
			data->threshold = t->threshold;
			data->rate = (t->rate == USE_DEFAULT) ? args.rate : t->rate;
			data->burst = (t->burst == USE_DEFAULT) ? args.burst : t->burst;
			data->decay = (t->decay == USE_DEFAULT) ? cur + args.decay : cur + t->decay;
			if (args.debug >= DEBUG_DUMP2) {
				inet_ntop(data->ss.ss_family, get_in_addr(&(data->ss)), buf2, sizeof(buf2));
				syslog(LOG_INFO, "setting %s to %u (%ld - %ld) = %ld", buf2, data->threshold, data->decay, cur, data->decay - cur);
			}
		}
		t = t->next;
	}

	/* Handle threshold changes */
	if (prev_threshold != data->threshold) {
		if (data->threshold > 0) {
			snprintf(buf, sizeof(buf), "%u", data->rate);
		} else {
			strncpy(buf, "none", sizeof(buf));
		}

		inet_ntop(data->ss.ss_family, get_in_addr(&(data->ss)), buf2, sizeof(buf2));
		syslog(LOG_INFO, "changing threshold for %s from %u to %u (pps %u, rate %s, burst %u, decay %ld)", buf2, prev_threshold, data->threshold, pps, buf, data->burst, (data->decay > 0) ? data->decay - cur : 0);

		if (data->threshold > prev_threshold) {
			snmptrap_threshold_increase(data, pps, prev_threshold);
		} else {
			snmptrap_threshold_decrease(data, pps, prev_threshold);
		}

		if (data->threshold > 0) {
			if (chain_add(data) == 0) { /* Handle failed insert */
				syslog(LOG_INFO, "failed to insert chain for %s - removing partial insert", buf2);

				chain_del(data);

				/* Clear crossed threshold */
				data->threshold = 0;
				data->rate = 0;
				data->burst = 0;
				data->decay = 0;
			}
		}

		if (prev_threshold > 0) {
			/* FIXME: We could save/restore using a memcpy with a tmp data_t struct. */
			/* Capture new threshold and rate */
			threshold = data->threshold;
			rate = data->rate;
			burst = data->burst;
			decay = data->decay;

			/* Use previous and delete rule. */
			data->threshold = prev_threshold;
			data->rate = prev_rate;
			data->burst = prev_burst;
			data->decay = prev_decay;
			chain_del(data);

			/* Restore new threshold and rate. */
			data->threshold = threshold;
			data->rate = rate;
			data->burst = burst;
			data->decay = decay;
		}
	}
}

int chain_add(struct data_t* data) {
	int i, ret;
	char cmd[4][512], buf[46];
	static uint16_t counter = 0;
	unsigned long index;

	if ((data->chain = malloc(MAX_CHAIN_LEN + 1)) == NULL) {
		syslog(LOG_ERR, "failed to alloc memory for chain name");

		return 0;
	}

	index = hash_ip(&(data->ss)) % HASH_WIDTH;
	snprintf(data->chain, MAX_CHAIN_LEN, "phreld-%08lx%04x%04x%04x", index, counter++, data->threshold % UINT16_MAX, pid % UINT16_MAX);
	inet_ntop(data->ss.ss_family, get_in_addr(&(data->ss)), buf, sizeof(buf));

	syslog(LOG_INFO, "inserting chain %s (%s)", data->chain, buf);

	/* Create new chain. */
	if (data->ss.ss_family == AF_INET) {
		snprintf(cmd[0], sizeof(cmd[0]), "%s -N %s > /dev/null 2>&1", args.iptables, data->chain);
	} else {
		snprintf(cmd[0], sizeof(cmd[0]), "%s -N %s > /dev/null 2>&1", args.ip6tables, data->chain);
	}

	/* Add rules to new chain. */
	if (data->rate > 0) {
		if (data->ss.ss_family == AF_INET) {
			snprintf(cmd[1], sizeof(cmd[1]), "%s -A %s -m limit --limit %u/sec --limit-burst %u -j RETURN > /dev/null 2>&1", args.iptables, data->chain, data->rate, data->burst);
		} else {
			snprintf(cmd[1], sizeof(cmd[1]), "%s -A %s -m limit --limit %u/sec --limit-burst %u -j RETURN > /dev/null 2>&1", args.ip6tables, data->chain, data->rate, data->burst);
		}
	} else {
		cmd[1][0] = '\0';
	}
	if (data->ss.ss_family == AF_INET) {
		snprintf(cmd[2], sizeof(cmd[2]), "%s -A %s -j %s > /dev/null 2>&1", args.iptables, data->chain, args.jump);
	} else {
		snprintf(cmd[2], sizeof(cmd[2]), "%s -A %s -j %s > /dev/null 2>&1", args.ip6tables, data->chain, args.jump);
	}

	/* Insert new chain into INPUT */
	if (data->ss.ss_family == AF_INET) {
		snprintf(cmd[3], sizeof(cmd[3]), "%s -I INPUT -j %s -s %s > /dev/null 2>&1", args.iptables, data->chain, buf);
	} else {
		snprintf(cmd[3], sizeof(cmd[3]), "%s -I INPUT -j %s -s %s > /dev/null 2>&1", args.ip6tables, data->chain, buf);
	}

	/* Loop though commands running them. */
	for (i = 0; i < 4; i++) {
		/* Skip empty commands. */
		if (cmd[i][0] == '\0') {
			continue;
		}

		if (args.debug >= DEBUG_INTERNAL) {
			if (args.test == 0) {
				syslog(LOG_INFO, "exec: %s", cmd[i]);
			} else {
				syslog(LOG_INFO, "exec: %s [test mode]", cmd[i]);
			}
		}

		/* Are we in test mode? */
		if (args.test == 0) {
			ret = system(cmd[i]); /* FIXME: Use exec() instead. */
		} else {
			ret = system(""); /* FIXME: Use exec() instead. */
		}

		/* Was the command successful? */
		if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)) {
			syslog(LOG_ERR, "failed to execute command: %s", cmd[i]);

			return 0;
		}

		if (WEXITSTATUS(ret) != 0) {
			syslog(LOG_ERR, "command not successful: %s", cmd[i]);

			return 0;
		}
	}

	/* Success */
	return 1;
}

int chain_del(struct data_t* data) {
	int i, ret;
	char cmd[3][512], buf[46];

	inet_ntop(data->ss.ss_family, get_in_addr(&(data->ss)), buf, sizeof(buf));

	syslog(LOG_INFO, "deleting chain %s (%s)", data->chain, buf);

	if (data->ss.ss_family == AF_INET) {
		/* Delete chain from Input */
		snprintf(cmd[0], sizeof(cmd[0]), "%s -D INPUT -j %s -s %s > /dev/null 2>&1", args.iptables, data->chain, buf);

		/* Flush Chain */
		snprintf(cmd[1], sizeof(cmd[1]), "%s -F %s > /dev/null 2>&1", args.iptables, data->chain);

		/* Delete Chain */
		snprintf(cmd[2], sizeof(cmd[2]), "%s -X %s > /dev/null 2>&1", args.iptables, data->chain);
	} else {
		/* Delete chain from Input */
		snprintf(cmd[0], sizeof(cmd[0]), "%s -D INPUT -j %s -s %s > /dev/null 2>&1", args.ip6tables, data->chain, buf);

		/* Flush Chain */
		snprintf(cmd[1], sizeof(cmd[1]), "%s -F %s > /dev/null 2>&1", args.ip6tables, data->chain);

		/* Delete Chain */
		snprintf(cmd[2], sizeof(cmd[2]), "%s -X %s > /dev/null 2>&1", args.ip6tables, data->chain);
	}

	/* Loop though commands running them. */
	for (i = 0; i < 3; i++) {
		/* Skip empty commands. */
		if (cmd[i][0] == '\0') {
			continue;
		}

		if (args.debug >= DEBUG_INTERNAL) {
			if (args.test == 0) {
				syslog(LOG_INFO, "exec: %s", cmd[i]);
			} else {
				syslog(LOG_INFO, "exec: %s [test mode]", cmd[i]);
			}
		}

		/* Are we in test mode? */
		if (args.test == 0) {
			ret = system(cmd[i]); /* FIXME: Use exec() instead. */
		} else {
			ret = system(""); /* FIXME: Use exec() instead. */
		}

		/* Was the command successful? */
		if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)) {
			syslog(LOG_ERR, "failed to execute command: %s", cmd[i]);

			return 0;
		}

		if (WEXITSTATUS(ret) != 0) {
			syslog(LOG_ERR, "command not successful: %s", cmd[i]);
		}
	}

	/* Success */
	return 1;
}

int chain_sync_add(void) {
	int i, ret;
	char cmd[2][512], chain[MAX_CHAIN_LEN];

	snprintf(chain, sizeof(chain), "phreld-sync-check-%04x", pid % UINT16_MAX);

	syslog(LOG_INFO, "inserting chain %s", chain);

	/* Create new chain. */
	if (args.ipv4) {
		snprintf(cmd[0], sizeof(cmd[0]), "%s -N %s > /dev/null 2>&1", args.iptables, chain);
	} else {
		cmd[0][0] = '\0';
	}

	if (args.ipv6) {
		snprintf(cmd[1], sizeof(cmd[1]), "%s -N %s > /dev/null 2>&1", args.ip6tables, chain);
	} else {
		cmd[1][0] = '\0';
	}

	/* Loop though commands running them. */
	for (i = 0; i < 2; i++) {
		/* Skip empty commands. */
		if (cmd[i][0] == '\0') {
			continue;
		}

		if (args.debug >= DEBUG_INTERNAL) {
			if (args.test == 0) {
				syslog(LOG_INFO, "exec: %s", cmd[i]);
			} else {
				syslog(LOG_INFO, "exec: %s [test mode]", cmd[i]);
			}
		}

		/* Are we in test mode? */
		if (args.test == 0) {
			ret = system(cmd[i]); /* FIXME: Use exec() instead. */
		} else {
			ret = system(""); /* FIXME: Use exec() instead. */
		}

		/* Was the command successful? */
		if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)) {
			syslog(LOG_ERR, "failed to execute command: %s", cmd[i]);

			return 0;
		}

		if (WEXITSTATUS(ret) != 0) {
			syslog(LOG_ERR, "command not successful: %s", cmd[i]);
		}
	}

	/* Success */
	return 1;
}

int chain_sync_check(void) {
	int i, ret;
	char cmd[2][512], chain[MAX_CHAIN_LEN];

	snprintf(chain, sizeof(chain), "phreld-sync-check-%04x", pid % UINT16_MAX);

	if (args.debug >= DEBUG_INTERNAL) {
		syslog(LOG_INFO, "checking chain %s", chain);
	}

	/* Create new chain. */
	if (args.ipv4) {
		snprintf(cmd[0], sizeof(cmd[0]), "%s -L %s > /dev/null 2>&1", args.iptables, chain);
	} else {
		cmd[0][0] = '\0';
	}

	if (args.ipv6) {
		snprintf(cmd[1], sizeof(cmd[1]), "%s -L %s > /dev/null 2>&1", args.ip6tables, chain);
	} else {
		cmd[1][0] = '\0';
	}

	/* Loop though commands running them. */
	for (i = 0; i < 2; i++) {
		/* Skip empty commands. */
		if (cmd[i][0] == '\0') {
			continue;
		}

		if (args.debug >= DEBUG_INTERNAL) {
			if (args.test == 0) {
				syslog(LOG_INFO, "exec: %s", cmd[i]);
			} else {
				syslog(LOG_INFO, "exec: %s [test mode]", cmd[i]);
			}
		}

		/* Are we in test mode? */
		if (args.test == 0) {
			ret = system(cmd[i]); /* FIXME: Use exec() instead. */
		} else {
			ret = system(""); /* FIXME: Use exec() instead. */
		}

		/* Was the command successful? */
		if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)) {
			syslog(LOG_ERR, "failed to execute command: %s", cmd[i]);

			return 0;
		}

		if (WEXITSTATUS(ret) != 0) {
			syslog(LOG_ERR, "command not successful: %s", cmd[i]);
		}
	}

	/* Success */
	return 1;
}

int chain_sync_del(void) {
	int i, ret;
	char cmd[2][512], chain[MAX_CHAIN_LEN];

	snprintf(chain, sizeof(chain), "phreld-sync-check-%04x", pid % UINT16_MAX);

	syslog(LOG_INFO, "deleting chain %s", chain);

	/* Delete Chain */
	if (args.ipv4) {
		snprintf(cmd[0], sizeof(cmd[0]), "%s -X %s > /dev/null 2>&1", args.iptables, chain);
	} else {
		cmd[0][0] = '\0';
	}

	if (args.ipv6) {
		snprintf(cmd[1], sizeof(cmd[1]), "%s -X %s > /dev/null 2>&1", args.ip6tables, chain);
	} else {
		cmd[1][0] = '\0';
	}

	/* Loop though commands running them. */
	for (i = 0; i < 2; i++) {
		/* Skip empty commands. */
		if (cmd[i][0] == '\0') {
			continue;
		}

		if (args.debug >= DEBUG_INTERNAL) {
			if (args.test == 0) {
				syslog(LOG_INFO, "exec: %s", cmd[i]);
			} else {
				syslog(LOG_INFO, "exec: %s [test mode]", cmd[i]);
			}
		}

		/* Are we in test mode? */
		if (args.test == 0) {
			ret = system(cmd[i]); /* FIXME: Use exec() instead. */
		} else {
			ret = system(""); /* FIXME: Use exec() instead. */
		}

		/* Was the command successful? */
		if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)) {
			syslog(LOG_ERR, "failed to execute command: %s", cmd[i]);

			return 0;
		}

		if (WEXITSTATUS(ret) != 0) {
			syslog(LOG_ERR, "command not successful: %s", cmd[i]);
		}
	}

	/* Success */
	return 1;
}

int calc_pps(struct data_t* data) {
	unsigned int i, ret_val = 0, sum = 0, max = 0;
	time_t cur = time(NULL), last_update = 0;
	char buf[46];

	if (args.debug >= DEBUG_DUMP2) {
		inet_ntop(data->ss.ss_family, get_in_addr(&(data->ss)), buf, sizeof(buf));
		syslog(LOG_INFO, "calculating pps for %s", buf);
	}

	for (i = 0; i < args.max_samples; i++) {
		/* Only count sample if it was updated within interval seconds. */
		if (data->last_update[i] + args.interval >= cur && data->last_update[i] != cur) {
			if (args.debug >= DEBUG_DUMP2) {
				/* Used data. Was updated within interval seconds. */
				syslog(LOG_INFO, "+sample[%02u] = %u (last: %ld sec)", i, data->samples[i], (data->last_update[i] > 0) ? cur - data->last_update[i]: 0);
			}

			if (data->samples[i] > max) {
				max = data->samples[i];
			}

			if (data->last_update[i] > last_update) {
				last_update = data->last_update[i];
			}

			sum += data->samples[i];
		} else {
			if (args.debug >= DEBUG_DUMP2) {
				/* Unused data. Too old... was not updated within last interval seconds. */
				syslog(LOG_INFO, " sample[%02u] = %u (last: %ld sec [expired])", i, data->samples[i], (data->last_update[i] > 0) ? cur - data->last_update[i]: 0);
			}
		}
	}

	data->local.last_update = last_update;
	data->local.avg = sum / args.interval;
	data->local.sum = sum;
	data->local.max = max;
	data->local.peak = (data->local.avg > data->local.peak) ? data->local.avg : data->local.peak;

	switch (args.algo) {
		case PPS_ALGO_AVG:
			ret_val = data->local.avg;
			break;

		case PPS_ALGO_SUM:
			ret_val = data->local.sum;
			break;

		case PPS_ALGO_MAX:
			ret_val = data->local.max;
			break;

		default:
			ret_val = 0;
	}

	if (data->db.last_update + args.stale_interval >= cur) {
		switch (args.db_algo) {
			case PPS_ALGO_AVG:
				ret_val = (data->local.avg + data->db.avg) / 2;
				break;

			case PPS_ALGO_SUM:
				ret_val = data->local.sum + data->db.sum;
				break;

			case PPS_ALGO_MAX:
				ret_val = (data->local.max > data->db.max) ? data->local.max : data->db.max;
				break;
		}
	}

	if (args.debug >= DEBUG_DUMP1) {
		inet_ntop(data->ss.ss_family, get_in_addr(&(data->ss)), buf, sizeof(buf));
		syslog(LOG_INFO, "%s has %u %s PPS (%u local packets in %u sec)", buf, ret_val, algo_names[args.algo], sum, args.interval);
	}

	return ret_val;
}

/*
	Create New:
	/sbin/iptables -N phreld-207.205.66.55-45

	Delete Chain
	/sbin/iptables -X phreld-207.205.66.55-45

	Append Rules onto existing chain (drop all):
	/sbin/iptables -A phreld-207.205.66.55-45 -j DROP

	Append Rules onto existing chain (rate limit):
	/sbin/iptables -A phreld-207.205.66.55-45 -p icmp -m limit --limit 600/min -j RETURN
	/sbin/iptables -A phreld-207.205.66.55-45 -p icmp -j DROP

	Insert chain into INPUT
	/sbin/iptables -I INPUT -j phreld-207.205.66.55-45

	Remove chain from INPUT
	/sbin/iptables -D INPUT -j phreld-207.205.66.55-45

*/

/*
** Local Variables:
** c-basic-offset: 3
** tab-width: 3
** End:
** vim: noet ts=3 sw=3
*/
